use std::path::Path;
use uuid::Uuid;
use proxmox_api::nodes::node::lxc;
use proxmox_api::nodes::node::lxc::vmid;
use proxmox_api::types::VmId;
use sea_orm::*;
use crate::api::cluster::node::wait_upid;
use crate::api::entity::locks::lock_vmid;
use crate::api::entity::{locks, nodes};
use crate::api::entity::nodes::P5xError;
use crate::api::entity::oplogs::Op;
use crate::api::services::{ssh_run_trimmed, Services};
use crate::api::services::SshError;


/** Create an empty Proxmox LXC container that can be used to shuttle a volume between PVE nodes. */
pub async fn provision_carrier(
    svc: &Services<'_>,
    from_node: &nodes::Model,
) -> Result<nodes::Model, P5xError> {
    // Make sure the empty filesystem template exists
    ensure_carrier_template(svc, &from_node.pve_host).await?;

    // Get the next VMID to be used by the carrier container
    let _lock = lock_vmid(svc.db, Some("Provisioning carrier container"))
        .await.map_err(P5xError::DbErr)?;

    let (hostname, vm_id) = svc.op_wrap_f(
        svc.op_start(Op::ProvisionCarrier, &format!("Host {}", &from_node.pve_host)).await?,
        "Provisioned carrier",
        async || {
            provision_carrier_unmanaged(svc, &from_node.pve_host).await
        },
    ).await?;

    // Create a new node instance
    let node = nodes::ActiveModel {
        hostname: Set(hostname),
        pve_id: Set(i32::try_from(vm_id.get()).unwrap()),
        pve_host: Set(from_node.pve_host.to_string()),
        assigned_ip: Set("0.0.0.0".to_string()),
        assigned_subnet: Set(0),
        is_permanent: Set(false),
        ..Default::default()
    }
        .save(svc.db)
        .await
        .and_then(|v| v.try_into_model())
        .map_err(P5xError::DbErr)?;

    // Create a lock instance for the new node
    locks::ActiveModel {
        lock_type: Set("nodes".to_string()),
        lock_resource: Set(node.id.to_string()),
        ..Default::default()
    }
        .save(svc.db)
        .await
        .map_err(P5xError::DbErr)?;

    Ok(node)
}


/** Like provision_carrier, but without any of the SQL bookkeeping, locking, or safety checks. */
pub async fn provision_carrier_unmanaged(
    svc: &Services<'_>,
    pve_host: &str,
) -> Result<(String, VmId), P5xError> {
    let pve_client = svc.pve().map_err(P5xError::ServiceError)?;
    let pve = proxmox_api::cluster::ClusterClient::new(pve_client);

    let vm_id = pve.nextid()
        .get(Default::default())
        .map_err(P5xError::PveError)?;
    let vm_id = VmId::new(vm_id).unwrap();

    // Build the new container params
    let hostname = format!("carrier-{}", Uuid::new_v4().to_string());
    let storage = &svc.config.pve_storage_pool;

    let mut params = lxc::PostParams::new("local:vztmpl/p5x-empty.tar.xz".to_string(), vm_id);
    params.cores = Some(1);
    params.description = Some("Temporary container managed by P5x".to_string());
    params.hostname = Some(hostname.to_string());
    params.memory = Some(16);  // in MB, min 16
    params.start = Some(false);
    params.storage = Some(storage.to_string());
    params.tags = Some("p5x".to_string());

    // Ask the PVE API to start creating the carrier node based on our empty template
    let upid = svc.pve_node(&pve_host)
        .map_err(P5xError::ServiceError)?
        .lxc()
        .post(params)
        .map_err(P5xError::PveError)?;

    // Wait for the container creation task to finish
    wait_upid(svc, &pve_host, &upid).await?;

    Ok((hostname, vm_id))
}

/** Make sure the tarball LXC template for the carrier container exists on the given node. */
async fn ensure_carrier_template(
    svc: &Services<'_>,
    pve_host: &str,
) -> Result<(), P5xError> {
    let pve_ssh = svc.pve_ssh(pve_host)
        .map_err(P5xError::ServiceError)?;

    // Use SFTP to check whether the file exists
    let path = Path::new("/var/lib/vz/template/cache/p5x-empty.tar.xz");
    let exists = pve_ssh.sftp()
        .and_then(|f| f.stat(path))
        .map_err(SshError::ClientError)
        .map_err(P5xError::SshError)?;

    // The empty template already exists, so we're done
    if exists.is_file() {
        return Ok(());
    }

    // Initialize the template directly
    // todo: we should really upload this to our p5x.image-cache during standup
    ssh_run_trimmed(&pve_ssh, "mkdir -p /tmp/p5x-empty-tmp")?;
    ssh_run_trimmed(&pve_ssh, "touch /tmp/p5x-empty-tmp/placeholder.txt")?;
    ssh_run_trimmed(&pve_ssh, "sh -c 'cd /tmp/p5x-empty-tmp && tar cfJ /var/lib/vz/template/cache/p5x-empty.tar.xz placeholder.txt'")?;
    ssh_run_trimmed(&pve_ssh, "rm -rf /tmp/p5x-empty-tmp")?;

    Ok(())
}


/** Destroy the given carrier LXC container. */
pub async fn terminate_carrier(
    svc: &Services<'_>,
    carrier: nodes::Model,
) -> Result<(), P5xError> {
    // pre-condition: the carrier is offline
    if !carrier.is_offline(svc)? {
        return Err(P5xError::BadPrecondition("Cannot terminate carrier instance: carrier is still running"));
    }

    // pre-condition: the carrier has a clean disk config
    if carrier.config(svc)?.has_pending() {
        return Err(P5xError::BadPrecondition("Cannot terminate carrier instance: carrier has unclean PVE config"))
    }

    svc.op_wrap_f(
        svc.op_start(Op::TerminateCarrier, &format!("Carrier {} ({})", carrier.vm_id(), &carrier.pve_host)).await?,
        "Terminated carrier",
        async || {
            terminate_carrier_unmanaged(svc, &carrier.pve_host, carrier.vm_id()).await?;

            nodes::Entity::delete_by_id(carrier.id)
                .exec(svc.db)
                .await
                .map_err(P5xError::DbErr)?;

            locks::Entity::delete_many()
                .filter(locks::Column::LockType.eq("nodes"))
                .filter(locks::Column::LockResource.eq(carrier.id.to_string()))
                .exec(svc.db)
                .await
                .map_err(P5xError::DbErr)?;

            Ok(())
        }
    ).await
}


/** Like terminate_carrier, but without any of the SQL bookkeeping, locking, or safety checks. */
pub async fn terminate_carrier_unmanaged(
    svc: &Services<'_>,
    pve_host: &str,
    vm_id: VmId,
) -> Result<(), P5xError> {
    let mut params = vmid::DeleteParams::default();
    params.purge = Some(true);
    params.destroy_unreferenced_disks = Some(true);

    let upid = svc.pve_node(pve_host)
        .map_err(P5xError::ServiceError)?
        .lxc()
        .vmid(vm_id)
        .delete(params)
        .map_err(P5xError::PveError)?;

    // post-condition: the carrier was successfully deleted
    wait_upid(svc, pve_host, &upid).await
}
