use std::{env, fs};
use std::fs::{File, Permissions};
use std::io::Write;
use std::os::unix::fs::PermissionsExt;
use std::path::Path;
use std::time::Duration;
use kube::{Api, Client};
use log::{info, debug};
use rand::rngs::OsRng;
use ssh_key::{PrivateKey, Algorithm, LineEnding};
use k8s_openapi::api::core::v1::{ConfigMap, Node, Pod};
use kube::api::{Patch, PatchParams};
use proxmox_api::nodes::node::lxc::vmid::move_volume;
use proxmox_api::types::VmId;
use proxmox_api::UreqError;
use serde_json::json;
use tokio::time::sleep;
use gethostname::gethostname;
use crate::api::cluster::carrier::{provision_carrier_unmanaged, terminate_carrier_unmanaged};
use crate::api::cluster::node::migrate_node_unmanaged;
use crate::api::cluster::volume::create_volume_unmanaged;
use crate::api::entity::nodes::{P5xError, PveConfig};
use crate::api::services::Services;

pub async fn ensure_system_disk(svc: &Services<'_>) -> Result<(), P5xError> {
    info!(target: "p5x", "Ensuring that the P5x API system disk exists...");

    // Load the dynamic-kv
    let client = Client::try_default().await.map_err(P5xError::KubeError)?;
    let namespace = fs::read_to_string("/var/run/secrets/kubernetes.io/serviceaccount/namespace")
        .or_else(|_| env::var("P5X_OVERRIDE_NS"))
        .unwrap_or_else(|_| "p5x-system".to_string());

    let maps: Api<ConfigMap> = Api::namespaced(client.clone(), &namespace);
    let map = maps.get("dynamic-kv").await.map_err(P5xError::KubeError)?;

    // If there's already a PVE host and disk mount in the config, we're done
    let mut data = map.data.unwrap_or_default();
    if let Some(host) = data.get("api-pve-host") {
        if let Some(mount) = data.get("api-pve-disk") {
            info!(target: "p5x", "Found existing P5x API system disk on {host}:{mount}");
            return Ok(());
        }
    }

    // Otherwise, create the disk on this PVE host
    info!(target: "p5x", "Provisioning new P5x API system disk (this is a one-time fixup)...");

    // Load the labels for this pod's node
    let pod_name = env::var("POD_NAME")
        .or_else(|_| gethostname().into_string())
        .expect("Could not determine pod name from environment");
    let pods: Api<Pod> = Api::namespaced(client.clone(), &namespace);
    let pod = pods.get(&pod_name).await.map_err(P5xError::KubeError)?;

    let node_name = pod.spec.and_then(|spec| spec.node_name).expect("Could not determine the Node name for pod!");

    let nodes: Api<Node> = Api::all(client);
    let node = nodes.get(&node_name).await.map_err(P5xError::KubeError)?;
    let labels = node.metadata.labels.expect("Could not load labels for node");

    let pve_host = labels.get("p5x.garrettmills.dev/pve-host")
        .expect("Node is missing required label: p5x.garrettmills.dev/pve-host");

    let pve_id: i32 = labels.get("p5x.garrettmills.dev/pve-id")
        .expect("Node is missing required label: p5x.garrettmills.dev/pve-id")
        .parse()
        .unwrap();

    debug!(target: "p5x", "pod {pod_name} | node {node_name} | host {pve_host} | vmid {pve_id}");

    // Create a new disk and mount it to the K8s node (LXC container)
    let (disk_name, _) = create_volume_unmanaged(
        svc,
        5 * 1024 * 1024 * 1024,
        pve_host,
        pve_id,
        "system-data"
    ).await?;

    // Add it to the dynamic-kv config and save
    data.insert("api-pve-host".to_string(), pve_host.to_string());
    data.insert("api-pve-id".to_string(), pve_id.to_string());
    data.insert("api-pve-disk".to_string(), disk_name);
    let patch = json!({
        "data": data
    });

    let params = PatchParams::apply("p5x-api");
    maps.patch("dynamic-kv", &params, &Patch::Merge(&patch))
        .await.map_err(P5xError::KubeError)?;

    info!(target: "p5x", "Successfully provisioned P5x API system disk.");
    Ok(())
}


pub async fn migrate_system_disk_if_necessary(svc: &Services<'_>) -> Result<(), P5xError> {
    // Load the dynamic-kv and get the current host/mount
    let client = Client::try_default().await.map_err(P5xError::KubeError)?;
    let namespace = fs::read_to_string("/var/run/secrets/kubernetes.io/serviceaccount/namespace")
        .or_else(|_| env::var("P5X_OVERRIDE_NS"))
        .unwrap_or_else(|_| "p5x-system".to_string());

    let maps: Api<ConfigMap> = Api::namespaced(client.clone(), &namespace);
    let map = maps.get("dynamic-kv").await.map_err(P5xError::KubeError)?;

    let data = map.data.unwrap_or_default();
    let current_host = data.get("api-pve-host").expect("Could not find api-pve-host in dynamic-kv config");
    let current_mount = data.get("api-pve-disk").expect("Could not find api-pve-disk in dynamic-kv config");
    let current_pve_id: i32 = data.get("api-pve-id").expect("Could not find api-pve-id in dynamic-kv config").parse().unwrap();

    // Load the labels for this pod's node
    let pod_name = env::var("POD_NAME")
        .or_else(|_| gethostname().into_string())
        .expect("Could not determine pod name from environment");
    let pods: Api<Pod> = Api::namespaced(client.clone(), &namespace);
    let pod = pods.get(&pod_name).await.map_err(P5xError::KubeError)?;

    let node_name = pod.spec.and_then(|spec| spec.node_name).expect("Could not determine the Node name for pod!");

    let nodes: Api<Node> = Api::all(client);
    let node = nodes.get(&node_name).await.map_err(P5xError::KubeError)?;
    let labels = node.metadata.labels.expect("Could not load labels for node");

    let pve_host = labels.get("p5x.garrettmills.dev/pve-host")
        .expect("Node is missing required label: p5x.garrettmills.dev/pve-host");

    let pve_id: i32 = labels.get("p5x.garrettmills.dev/pve-id")
        .expect("Node is missing required label: p5x.garrettmills.dev/pve-id")
        .parse()
        .unwrap();

    // If the disk is already on our LXC container (K8s node), we're done
    if pve_id == current_pve_id {
        return Ok(())
    }

    // Otherwise, we need to migrate the disk to this node
    // FIXME: Consideration: What if the disk is still mounted somewhere?
    transfer_unmanaged(
        svc, current_mount, current_host, current_pve_id, pve_host, pve_id).await?;

    Ok(())
}


/** Check if the SSH pubkey/privkey exist at the configured paths. If not, generate them. */
pub fn ensure_ssh_keypair() -> Result<(), ssh_key::Error> {
    let pubkey_path = env::var("P5X_SSH_PUBKEY_PATH").expect("Missing env: P5X_SSH_PUBKEY_PATH");
    let privkey_path = env::var("P5X_SSH_PRIVKEY_PATH").expect("Missing env: P5X_SSH_PRIVKEY_PATH");

    // If both exist, then p5x will boot correctly when it reads from the files.
    // If only one of the two files exists, p5x will error on boot. This is safer
    // than accidentally overwriting someone's key.
    if Path::new(&pubkey_path).exists() || Path::new(&privkey_path).exists() {
        info!(target: "p5x", "Found existing SSH keypair on filesystem");
        return Ok(());
    }

    // Generate an ed25519 keypair
    info!(target: "p5x", "Generating a new SSH keypair");
    let mut csprng = OsRng;
    let privkey = PrivateKey::random(&mut csprng, Algorithm::Ed25519)?;

    // Write the privkey to a file
    let privkey_pem = privkey.to_openssh(LineEnding::LF)?;
    let mut privkey_file = File::create(privkey_path)?;
    privkey_file.write_all(privkey_pem.as_bytes())?;
    privkey_file.set_permissions(Permissions::from_mode(0o600))?;

    // Write the pubkey to a file
    let pubkey = privkey.public_key();
    let pubkey_ssh = pubkey.to_openssh()?;
    let mut pubkey_file = File::create(pubkey_path)?;
    pubkey_file.write_all(pubkey_ssh.as_bytes())?;

    Ok(())
}


// todo: these are very similar to transfer() and transfer_directly(), but there was too much
//       sql bookkeeping in those to generalize effectively.

/** Migrate a volume from its current LXC container to the specified LXC container. */
async fn transfer_unmanaged(
    svc: &Services<'_>,
    old_mountpoint_identifier: &str,
    from_pve_host: &str,
    from_pve_id: i32,
    to_pve_host: &str,
    to_pve_id: i32,
) -> Result<String, P5xError> {
    // If from_node and to_node are on the same physical host, transfer directly
    if from_pve_host == to_pve_host {
        return transfer_directly_unmanaged(
            svc, old_mountpoint_identifier, from_pve_host, from_pve_id, to_pve_host, to_pve_id).await;
    }

    // If the nodes are on different physical hosts, we need to create a temporary
    // container on shared storage to attach the volume to. We'll then migrate that
    // container to the target physical host.
    let (hostname, vm_id) = provision_carrier_unmanaged(svc, from_pve_host).await?;
    let carrier_mountpoint = transfer_directly_unmanaged(
        svc, old_mountpoint_identifier, from_pve_host, from_pve_id, &hostname, vm_id.get() as i32).await?;

    // Migrate the carrier to the target host
    migrate_node_unmanaged(svc, from_pve_host, vm_id.get() as i32, to_pve_host).await?;

    // Transfer the volume from the carrier to the target LXC container
    let new_mountpoint = transfer_directly_unmanaged(
        svc, &carrier_mountpoint, &hostname, vm_id.get() as i32, to_pve_host, to_pve_id).await?;

    // Delete the carrier container
    terminate_carrier_unmanaged(svc, to_pve_host, vm_id).await?;

    Ok(new_mountpoint)
}


async fn transfer_directly_unmanaged(
    svc: &Services<'_>,
    old_mountpoint_identifier: &str,
    from_pve_host: &str,
    from_pve_id: i32,
    to_pve_host: &str,
    to_pve_id: i32,
) -> Result<String, P5xError> {
    // pre-condition: to_node's config is clean
    let old_to_config = PveConfig::load(svc, to_pve_host, to_pve_id)?;
    if old_to_config.has_pending() {
        return Err(P5xError::BadPrecondition("Cannot transfer volume: to-node has unclean LXC config"))
    }

    // pre-condition: from_node's config is clean
    let old_from_config = PveConfig::load(svc, from_pve_host, from_pve_id)?;
    if old_from_config.has_pending() {
        return Err(P5xError::BadPrecondition("Cannot transfer volume: from-node has unclean LXC config"))
    }

    // Figure out the mountpoint identifier on the new node
    let new_mountpoint_identifier = old_to_config.next_nth("unused");
    let new_mountpoint_identifier = format!("unused{new_mountpoint_identifier}");

    // Ask the PVE API to move the volume to the new node
    let pve_from_vol = move_volume::Volume::try_from(old_mountpoint_identifier).unwrap();
    let pve_to_vol = move_volume::TargetVolume::try_from(new_mountpoint_identifier.as_str()).unwrap();
    let mut post_params = move_volume::PostParams::new(pve_from_vol);
    post_params.target_volume = Some(pve_to_vol);
    post_params.target_vmid = Some(VmId::new(i64::from(to_pve_id)).unwrap());

    let res = svc.pve_node(from_pve_host)
        .map_err(P5xError::ServiceError)?
        .lxc()
        .vmid(VmId::new(i64::from(from_pve_id)).unwrap())
        .move_volume()
        .post(post_params);

    // This is necessary because POST returns {data: null} on success,
    // which the UreqClient flags as an unknown error.
    if let Err(UreqError::EncounteredErrors(e)) = res {
        return Err(P5xError::PveError(UreqError::EncounteredErrors(e)));
    }

    // Wait for the volume to transfer, since we don't get back a UPID
    sleep(Duration::from_secs(10)).await;

    // post-condition: from_node's config is clean
    let new_from_config = PveConfig::load(svc, from_pve_host, from_pve_id)?;
    if new_from_config.has_pending() {
        // todo: don't have a great way to revert this currently w/o a ton of work
        return Err(P5xError::BadPostcondition("Could not transfer volume: from-node had unclean LXC config after transfer"))
    }

    // post-condition: to_node's config is clean
    let new_to_config = PveConfig::load(svc, to_pve_host, to_pve_id)?;
    if new_to_config.has_pending() {
        // todo: don't have a great way to revert this currently w/o a ton of work
        return Err(P5xError::BadPostcondition("Could not transfer volume: to-node had unclean LXC config after transfer"))
    }

    // Verify that the volume appears in the new node's config
    new_to_config.get(&new_mountpoint_identifier)
        .ok_or(P5xError::BadPostcondition("Could not find mountpoint config after transferring volume"))?;

    Ok(new_mountpoint_identifier)
}
