use std::future::Future;
use std::io::Read;
use std::net::TcpStream;
use proxmox_api::nodes::node::NodeClient;
use proxmox_api::UreqClient;
use sea_orm::*;
use sea_orm::ActiveValue::Set;
use sea_orm::sqlx::types::time::OffsetDateTime;
use ssh2::Session;
use crate::api::entity::nodes::P5xError;
use crate::api::util::{read_p5x_config, P5xConfig};
use crate::api::entity::oplogs;
use crate::api::entity::oplogs::Op;

#[derive(Debug)]
pub enum SshError {
    IoError(std::io::Error),
    ClientError(ssh2::Error),
    CommandFailed(i32, String),
    AuthFailed,
}

impl SshError {
    pub fn command_failed(exit_code: i32, command: &str, output: &str) -> SshError {
        SshError::CommandFailed(exit_code, format!("Command `{command}` failed: {output}"))
    }
}

#[derive(Debug)]
pub enum ServiceError {
    PveError(proxmox_api::UreqError),
    InvalidNetworkInterface,
    SshError(SshError),
}


/** Helper for managing singleton services. */
pub struct Services<'a> {
    pub db: &'a DatabaseConnection,
    pub config: P5xConfig,
}

impl<'a> Services<'a> {
    /** Create a new services instance. */
    pub async fn build(db: &'a DatabaseConnection) -> Result<Services<'a>, DbErr> {
        Ok(Services {
            db,
            config: read_p5x_config(),
        })
    }

    /** Get a new Proxmox API client instance. */
    pub fn pve(&self) -> Result<UreqClient, ServiceError> {
        let host = &self.config.pve_api_host;
        let pw = &self.config.pve_root_password;

        let host = format!("https://{host}:8006");
        let api = UreqClient::new(&host, "root", "pam", &pw)
            .map_err(ServiceError::PveError)?;

        Ok(api)
    }

    /** Get a Proxmox Node API client instance for the given node. */
    pub fn pve_node(&self, host: &str) -> Result<NodeClient<UreqClient>, ServiceError> {
        let pve = self.pve()?;
        Ok(proxmox_api::nodes::NodesClient::new(pve).node(host))
    }

    /** Given a PVE host name, get its network address. */
    pub fn pve_addr(&self, host: &str) -> Result<String, ServiceError> {
        let ifaces = self.pve_node(host)?
            .network()
            .get(Default::default())
            .map_err(ServiceError::PveError)?;

        let addr = ifaces.iter()
            .filter_map(|i| i.additional_properties.get("address"))
            .next()
            .ok_or(ServiceError::InvalidNetworkInterface)?;

        if !addr.is_string() {
            return Err(ServiceError::InvalidNetworkInterface);
        }

        Ok(addr.as_str().unwrap().to_string())
    }

    /** Open a new SSH session to the given PVE node. */
    pub fn pve_ssh(&self, host: &str) -> Result<Session, ServiceError> {
        let addr = self.pve_addr(host)?;
        let addr = format!("{}:22", addr);
        let tcp = TcpStream::connect(addr)
            .map_err(SshError::IoError)
            .map_err(ServiceError::SshError)?;

        let mut sess = Session::new()
            .map_err(SshError::ClientError)
            .map_err(ServiceError::SshError)?;

        sess.set_tcp_stream(tcp);
        sess.handshake()
            .map_err(SshError::ClientError)
            .map_err(ServiceError::SshError)?;

        sess.userauth_password("root", &self.config.pve_root_password)
            .map_err(SshError::ClientError)
            .map_err(ServiceError::SshError)?;

        if !sess.authenticated() {
            return Err(ServiceError::SshError(SshError::AuthFailed));
        }

        Ok(sess)
    }

    pub async fn op_unwrap<Ret, Err>(&self, op: i64, val: Result<Ret, Err>) -> Result<Ret, P5xError>
    where
        Err: std::fmt::Debug,
        P5xError: From<Err>,
    {
        match val {
            Ok(v) => Ok(v),
            Err(e) => {
                self.op_fail(op, &format!("{:?}", e)).await?;
                Err(P5xError::from(e))
            }
        }
    }

    pub async fn op_unwrap_f<Ret, Err, FutureVal>(&self, op: i64, val: FutureVal) -> Result<Ret, P5xError>
    where
        Err: std::fmt::Debug,
        P5xError: From<Err>,
        FutureVal: Future<Output = Result<Ret, Err>>,
    {
        self.op_unwrap(op, val.await).await
    }

    pub async fn op_wrap_f<Ret, Closure, FutureClosure>(&self, op: i64, success: &str, cb: Closure) -> Result<Ret, P5xError>
    where
        FutureClosure: Future<Output = Result<Ret, P5xError>>,
        Closure: FnOnce() -> FutureClosure,
    {
        match cb().await {
            Ok(v) => {
                self.op_finish(op, success).await?;
                Ok(v)
            },
            Err(e) => {
                self.op_fail(op, &format!("{:?}", e)).await?;
                Err(e)
            }
        }
    }

    pub async fn op_start(&self, op: Op, desc: &str) -> Result<i64, P5xError> {
        let res = oplogs::ActiveModel {
            op: Set(op.to_string()),
            op_desc: Set(desc.to_string()),
            op_at: Set(OffsetDateTime::now_utc()),
            ..Default::default()
        }
            .save(self.db)
            .await
            .map_err(P5xError::DbErr)?
            .try_into_model()
            .map_err(P5xError::DbErr)?;

        Ok(res.id)
    }

    pub async fn op_finish(&self, id: i64, desc: &str) -> Result<(), P5xError> {
        self.op_result(true, id, desc).await
    }

    pub async fn op_fail(&self, id: i64, desc: &str) -> Result<(), P5xError> {
        self.op_result(false, id, desc).await
    }

    async fn op_result(&self, result: bool, id: i64, desc: &str) -> Result<(), P5xError> {
        let mut op = self.resolve_op(id).await?.into_active_model();
        op.result = Set(Some(result));
        op.result_desc = Set(Some(desc.to_string()));
        op.result_at = Set(Some(OffsetDateTime::now_utc()));
        op.save(self.db)
            .await
            .map_err(P5xError::DbErr)?;

        Ok(())
    }

    async fn resolve_op(&self, id: i64) -> Result<oplogs::Model, P5xError> {
        oplogs::Entity::find_by_id(id)
            .one(self.db)
            .await
            .map_err(P5xError::DbErr)?
            .ok_or(P5xError::BadPrecondition("Could not resolve Oplog model instance"))
    }
}


/** Run an SSH command and return the output as a string, with whitespace trimmed. */
pub fn ssh_run_trimmed(session: &Session, cmd: &str) -> Result<String, P5xError> {
    let mut channel = session.channel_session()
        .map_err(SshError::ClientError)
        .map_err(P5xError::SshError)?;

    channel.exec(cmd)
        .map_err(SshError::ClientError)
        .map_err(P5xError::SshError)?;

    let mut contents = String::new();
    channel.read_to_string(&mut contents)
        .map_err(SshError::IoError)
        .map_err(P5xError::SshError)?;

    let exit_code = channel.exit_status()
        .map_err(SshError::ClientError)
        .map_err(P5xError::SshError)?;

    let contents = contents.trim();
    if exit_code != 0 {
        return Err(P5xError::SshError(SshError::command_failed(exit_code, cmd, &contents)));
    }

    Ok(contents.to_string())
}
