pub mod api;
#[macro_use] extern crate rocket;
use dotenv::dotenv;
use rocket::{Build, Rocket};
use log::{error, info, debug};
use std::{env, process};
use sea_orm::Database;
use crate::api::cluster::system::{ensure_ssh_keypair, ensure_system_disk, migrate_system_disk_if_necessary};
use crate::api::services::Services;
use crate::api::util::read_p5x_config;
use crate::api::db::migrations::run_migrations_unmanaged;

fn configure_rocket() -> Rocket<Build> {
    rocket::build()
        .attach(api::init())
}

#[tokio::main]
async fn main() {
    dotenv().ok();
    env_logger::init();
    info!(target: "p5x", "Starting p5x...");

    let args: Vec<String> = env::args().collect();
    if args.len() < 2 {
        error!(target: "p5x", "Missing required <mode> argument. Valid modes: api-server,ensure-system-disk,migrate-system-disk");
        process::exit(1);
    }

    let mode = &args[1];
    debug!(target: "p5x", "Running with mode: {mode}");

    // Intentionally generate this before migrating/ensuring the system disk, since that requires
    // Services, and Services requires the SSH keys to exist. The keys generated during the system
    // disk ops will be overwritten when the disk is mounted.
    ensure_ssh_keypair().expect("Could not ensure SSH keypair exists.");

    if mode == "ensure-system-disk" {
        let anon_db = Database::connect("sqlite::memory:").await.unwrap();
        run_migrations_unmanaged(&anon_db).await.unwrap();
        let svc = Services::build(&anon_db).await.unwrap();
        ensure_system_disk(&svc).await.unwrap();
        return;
    }

    if mode == "migrate-system-disk" {
        let anon_db = Database::connect("sqlite::memory:").await.unwrap();
        run_migrations_unmanaged(&anon_db).await.unwrap();
        let svc = Services::build(&anon_db).await.unwrap();
        migrate_system_disk_if_necessary(&svc).await.unwrap();
        return;
    }

    let config = read_p5x_config();  // Do this so we early-fail if there are missing env vars
    info!(target: "p5x", "Successfully read config from environment.");
    info!(target: "p5x", "Cluster host: {} ({})", config.pve_host_name, config.pve_api_host);
    info!(target: "p5x", "Storage pool: {} ({})", config.pve_storage_pool, config.pve_storage_driver);

    if mode == "api-server" {
        let rocket = configure_rocket();
        if let Err(e) = rocket.launch().await {
            error!(target: "p5x", "Rocket failed to launch: {:?}", e)
        }
        return;
    }
}
